import logging
import os
import sys
from abc import ABC, abstractmethod
from collections import defaultdict, OrderedDict
from pprint import pformat

import tensorflow as tf
from tqdm import tqdm

from ludwig.constants import COMBINED, LOGITS, LAST_HIDDEN
from ludwig.globals import is_progressbar_disabled
from ludwig.utils.data_utils import flatten_df, from_numpy_dataset, save_json
from ludwig.utils.horovod_utils import initialize_horovod, return_first
from ludwig.utils.misc_utils import sum_dicts
from ludwig.utils.print_utils import repr_ordered_dict
from ludwig.utils.tf_utils import initialize_tensorflow

EXCLUE_PRED_SET = {LOGITS, LAST_HIDDEN}
SKIP_EVAL_METRICS = {'confusion_matrix', 'roc_curve'}

logger = logging.getLogger(__name__)


class BasePredictor(ABC):
    @abstractmethod
    def batch_predict(
            self,
            model,
            dataset,
            dataset_name=None
    ):
        raise NotImplementedError()

    @abstractmethod
    def batch_evaluation(
            self,
            model,
            dataset,
            collect_predictions=False,
            dataset_name=None
    ):
        raise NotImplementedError()

    @abstractmethod
    def batch_collect_activations(
            self,
            model,
            layer_names,
            dataset,
            bucketing_field=None
    ):
        raise NotImplementedError()

    # Remote implementations may override this
    def shutdown(self):
        pass

    # Functions needed to treat Trainer as a context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()


class Predictor(BasePredictor):
    """
    Predictor is a class that uses a model to predict and evaluate
    """

    def __init__(
            self,
            batch_size=128,
            horovod=None,
            debug=False,
            **kwargs
    ):
        self._batch_size = batch_size
        self._horovod = horovod
        self._debug = debug

    def batch_predict(
            self,
            model,
            dataset,
            dataset_name=None
    ):
        with dataset.initialize_batcher(
            self._batch_size,
            should_shuffle=False,
            horovod=self._horovod
        ) as batcher:

            progress_bar = None
            if self.is_coordinator():
                progress_bar = tqdm(
                    desc='Prediction' if dataset_name is None
                    else 'Prediction {0: <5.5}'.format(dataset_name),
                    total=batcher.steps_per_epoch,
                    file=sys.stdout,
                    disable=is_progressbar_disabled()
                )

            predictions = defaultdict(list)
            while not batcher.last_batch():
                batch = batcher.next_batch()

                inputs = {
                    i_feat.feature_name: batch[i_feat.proc_column]
                    for i_feat in model.input_features.values()
                }

                preds = model.predict_step(inputs)

                # accumulate predictions from batch for each output feature
                for of_name, of_preds in preds.items():
                    for pred_name, pred_values in of_preds.items():
                        if pred_name not in EXCLUE_PRED_SET:
                            key = f'{of_name}_{pred_name}'
                            predictions[key].append(pred_values)

                if self.is_coordinator():
                    progress_bar.update(1)

            if self.is_coordinator():
                progress_bar.close()

        # consolidate predictions from each batch to a single tensor
        for key, pred_value_list in predictions.items():
            predictions[key] = tf.concat(pred_value_list, axis=0).numpy()

        return from_numpy_dataset(predictions)

    def batch_evaluation(
            self,
            model,
            dataset,
            collect_predictions=False,
            dataset_name=None
    ):
        with dataset.initialize_batcher(
            self._batch_size,
            should_shuffle=False,
            horovod=self._horovod
        ) as batcher:

            progress_bar = None
            if self.is_coordinator():
                progress_bar = tqdm(
                    desc='Evaluation' if dataset_name is None
                    else 'Evaluation {0: <5.5}'.format(dataset_name),
                    total=batcher.steps_per_epoch,
                    file=sys.stdout,
                    disable=is_progressbar_disabled()
                )

            predictions = defaultdict(list)
            while not batcher.last_batch():
                batch = batcher.next_batch()

                inputs = {
                    i_feat.feature_name: batch[i_feat.proc_column]
                    for i_feat in model.input_features.values()
                }
                targets = {
                    o_feat.feature_name: batch[o_feat.proc_column]
                    for o_feat in model.output_features.values()
                }

                preds = model.evaluation_step(inputs, targets)

                # accumulate predictions from batch for each output feature
                if collect_predictions:
                    for of_name, of_preds in preds.items():
                        for pred_name, pred_values in of_preds.items():
                            if pred_name not in EXCLUE_PRED_SET:
                                key = f'{of_name}_{pred_name}'
                                predictions[key].append(pred_values)

                if self.is_coordinator():
                    progress_bar.update(1)

            if self.is_coordinator():
                progress_bar.close()

        # consolidate predictions from each batch to a single tensor
        if collect_predictions:
            for key, pred_value_list in predictions.items():
                predictions[key] = tf.concat(pred_value_list, axis=0).numpy()

        metrics = model.get_metrics()
        metrics = self.merge_workers_metrics(metrics)
        model.reset_metrics()

        return metrics, from_numpy_dataset(predictions)

    def batch_collect_activations(
            self,
            model,
            layer_names,
            dataset,
            bucketing_field=None
    ):
        if bucketing_field:
            raise ValueError('BucketedBatcher is not supported yet')

        # Build static graph for the trained model
        tf.keras.backend.reset_uids()
        keras_model_inputs = model.get_model_inputs(training=False)
        keras_model = model.get_connected_model(inputs=keras_model_inputs,
                                                training=False)

        # Create a new model that routes activations to outputs
        tf.keras.backend.reset_uids()
        output_nodes = {layer_name: keras_model.get_layer(layer_name).output
                        for layer_name in layer_names}
        activation_model = tf.keras.Model(inputs=keras_model_inputs,
                                          outputs=output_nodes)

        with dataset.initialize_batcher(
            self._batch_size,
            should_shuffle=False
        ) as batcher:

            progress_bar = tqdm(
                desc='Collecting Tensors',
                total=batcher.steps_per_epoch,
                file=sys.stdout,
                disable=is_progressbar_disabled()
            )

            collected_tensors = []
            while not batcher.last_batch():
                batch = batcher.next_batch()

                inputs = {
                    i_feat.feature_name: batch[i_feat.proc_column]
                    for i_feat in model.input_features.values()
                }
                outputs = activation_model(inputs)

                for layer_name, output in outputs.items():
                    if isinstance(output, tuple):
                        output = list(output)

                    if isinstance(output, tf.Tensor):
                        output = [('', output)]
                    elif isinstance(output, dict):
                        output = [(f'_{key}', tensor)
                                  for key, tensor in output.items()]
                    elif isinstance(output, list):
                        output = [(f'_{idx}', tensor)
                                  for idx, tensor in enumerate(output)]

                    for suffix, tensor in output:
                        full_name = f'{layer_name}{suffix}'
                        collected_tensors.append((full_name, tensor))

                progress_bar.update(1)

            progress_bar.close()

        return collected_tensors

    def merge_workers_metrics(self, metrics):
        if not self._horovod:
            return metrics

        # gather outputs from all workers
        all_workers_output_metrics = self._horovod.allgather_object(metrics)

        # merge them into a single one
        merged_output_metrics = sum_dicts(
            all_workers_output_metrics,
            dict_type=OrderedDict
        )

        return merged_output_metrics

    def is_coordinator(self):
        if not self._horovod:
            return True
        return self._horovod.rank() == 0


class RemotePredictor(Predictor):
    def __init__(
        self,
        gpus=None,
        gpu_memory_limit=None,
        allow_parallel_threads=True,
        **kwargs
    ):
        horovod = initialize_horovod()
        initialize_tensorflow(gpus=gpus,
                              gpu_memory_limit=gpu_memory_limit,
                              allow_parallel_threads=allow_parallel_threads,
                              horovod=horovod)
        super().__init__(horovod=horovod, **kwargs)

        # Only return results from rank 0 to reduce network overhead
        self.batch_predict = return_first(self.batch_predict)
        self.batch_evaluation = return_first(self.batch_evaluation)


def calculate_overall_stats(
        output_features,
        predictions,
        dataset,
        training_set_metadata
):
    overall_stats = {}
    for of_name, output_feature in output_features.items():
        feature_metadata = output_feature.overall_statistics_metadata()
        feature_metadata.update(
            training_set_metadata[output_feature.feature_name])

        feature_df = predictions.loc[:, predictions.columns.str.startswith(of_name)]
        feature_df = feature_df.rename(columns=lambda c: c[len(of_name) + 1:])

        overall_stats[of_name] = output_feature.calculate_overall_stats(
            feature_df,  # predictions
            dataset.get(output_feature.proc_column),  # target
            feature_metadata,  # output feature metadata
        )
    return overall_stats


def save_prediction_outputs(
        postprocessed_output,
        output_directory,
        backend,
):
    postprocessed_output, column_shapes = flatten_df(
        postprocessed_output, backend
    )
    postprocessed_output.to_parquet(
        os.path.join(output_directory, 'predictions.parquet')
    )
    save_json(
        os.path.join(output_directory, 'predictions.shapes.json'),
        column_shapes
    )


def save_evaluation_stats(test_stats, output_directory):
    test_stats_fn = os.path.join(
        output_directory,
        'test_statistics.json'
    )
    save_json(test_stats_fn, test_stats)


def print_evaluation_stats(test_stats):
    for output_field, result in test_stats.items():
        if (output_field != COMBINED or
                (output_field == COMBINED and len(test_stats) > 2)):
            logger.info('\n===== {} ====='.format(output_field))
            for metric in sorted(list(result)):
                if metric not in SKIP_EVAL_METRICS:
                    value = result[metric]
                    if isinstance(value, OrderedDict):
                        value_repr = repr_ordered_dict(value)
                    else:
                        value_repr = pformat(result[metric], indent=2)
                    logger.info('{0}: {1}'.format(metric, value_repr))


def get_output_columns(output_features):
    output_columns = []
    for of_name, feature in output_features.items():
        for pred in feature.get_prediction_set():
            if pred not in EXCLUE_PRED_SET:
                output_columns.append(f'{of_name}_{pred}')
    return output_columns



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2021 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import pandas as pd

from ludwig.datasets.base_dataset import BaseDataset, DEFAULT_CACHE_LOCATION
from ludwig.datasets.mixins.download import UncompressedFileDownloadMixin
from ludwig.datasets.mixins.load import CSVLoadMixin
from ludwig.datasets.mixins.process import MultifileJoinProcessMixin

def load(cache_dir=DEFAULT_CACHE_LOCATION, split=False):
    dataset = MushroomEdibility(cache_dir=cache_dir)
    return dataset.load(split=split)

class MushroomEdibility(UncompressedFileDownloadMixin, MultifileJoinProcessMixin,
            CSVLoadMixin, BaseDataset):
    """
    The Mushroom Edibility dataset

    Additional Details:

    http://archive.ics.uci.edu/ml/machine-learning-databases/mushroom/agaricus-lepiota.names
    """
    def __init__(self, cache_dir=DEFAULT_CACHE_LOCATION):
        super().__init__(dataset_name="mushroom_edibility", cache_dir=cache_dir)

    def process_downloaded_dataset(self):
        super().process_downloaded_dataset(header=None)
        processed_df = pd.read_csv(os.path.join(self.processed_dataset_path,
                                                self.csv_filename))
        columns = [
            "class", "cap-shape", "cap-surface", "cap-color", "bruises?", "odor",
            "gill-attachment", "gill-spacing", "gill-size", "gill-color",
            "stalk-shape", "stalk-root", "stalk-surface-above-ring",
            "stalk-surface-below-ring", "stalk-color-above-ring",
            "stalk-color-below-ring", "veil-type", "veil-color", "ring-number",
            "ring-type", "spore-print-color", "population", "habitat", "split"
        ]
        processed_df.columns = columns
        processed_df.to_csv(
            os.path.join(self.processed_dataset_path, self.csv_filename),
            index=False
        )
        


#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import gzip
import os
import shutil
import tarfile
import urllib.request
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile

from tqdm import tqdm

class TqdmUpTo(tqdm):
    """Provides progress bar for `urlretrieve`.

    Taken from: https://gist.github.com/leimao/37ff6e990b3226c2c9670a2cd1e4a6f5
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


class ZipDownloadMixin:
    """Downloads the zip file containing the training data and extracts the contents."""

    config: dict
    raw_dataset_path: str
    raw_temp_path: str

    def download_raw_dataset(self):
        """
        Download the raw dataset and extract the contents of the zip file and
        store that in the cache location.
        """
        os.makedirs(self.raw_temp_path, exist_ok=True)
        for url in self.download_urls:
            with urlopen(url) as zipresp:
                with ZipFile(BytesIO(zipresp.read())) as zfile:
                    zfile.extractall(self.raw_temp_path)
        os.rename(self.raw_temp_path, self.raw_dataset_path)

    @property
    def download_urls(self):
        return self.config["download_urls"]

class TarDownloadMixin:
    """Downloads the compressed tar file containing the training data and extracts the contents."""

    config: dict
    raw_dataset_path: str
    raw_temp_path: str

    def download_raw_dataset(self):
        """
        Download the raw dataset and extract the contents of the tar file and
        store that in the cache location.
        """
        os.makedirs(self.raw_temp_path, exist_ok=True)
        for url in self.download_urls:
            filename = url.split('/')[-1]
            with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024,
                          miniters=1, desc=filename) as t:
                urllib.request.urlretrieve(
                    url,
                    os.path.join(self.raw_temp_path, filename),
                    t.update_to
                )

            download_folder_name = url.split('/')[-1].split('.')[0]
            file_path = os.path.join(self.raw_temp_path, filename)
            with tarfile.open(file_path) as tar_file:
                tar_file.extractall(path=self.raw_temp_path)

            for f in os.scandir(os.path.join(self.raw_temp_path, 
                                             download_folder_name)):
                shutil.copyfile(f, os.path.join(self.raw_temp_path, f.name))

        os.rename(self.raw_temp_path, self.raw_dataset_path)

    @property
    def download_urls(self):
        return self.config['download_urls']

class GZipDownloadMixin:
    """Downloads the gzip archive file containing the training data and extracts the contents."""

    config: dict
    raw_dataset_path: str
    raw_temp_path: str

    def download_raw_dataset(self):
        """
        Download the raw dataset and extract the contents of the zip file and
        store that in the cache location.
        """
        os.makedirs(self.raw_temp_path, exist_ok=True)
        for file_download_url in self.download_urls:
            filename = file_download_url.split('/')[-1]
            with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024,
                          miniters=1, desc=filename) as t:
                urllib.request.urlretrieve(
                    file_download_url,
                    os.path.join(self.raw_temp_path, filename),
                    t.update_to
                )
            gzip_content_file = '.'.join(filename.split('.')[:-1])
            with gzip.open(
                    os.path.join(self.raw_temp_path, filename)) as gzfile:
                with open(os.path.join(self.raw_temp_path, gzip_content_file),
                          'wb') as output:
                    shutil.copyfileobj(gzfile, output)
        os.rename(self.raw_temp_path, self.raw_dataset_path)

    @property
    def download_urls(self):
        return self.config["download_urls"]


class UncompressedFileDownloadMixin:
    """Downloads the json file containing the training data and extracts the contents."""

    config: dict
    raw_dataset_path: str
    raw_temp_path: str

    def download_raw_dataset(self):
        """
        Download the raw dataset files and store in the cache location.
        """
        os.makedirs(self.raw_temp_path, exist_ok=True)
        for url in self.download_url:
            filename = url.split('/')[-1]
            urllib.request.urlretrieve(url, os.path.join(self.raw_temp_path,
                                                         filename))

        os.rename(self.raw_temp_path, self.raw_dataset_path)

    @property
    def download_url(self):
        return self.config["download_urls"]



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import struct
from multiprocessing.pool import ThreadPool

import numpy as np
from skimage.io import imsave

from ludwig.datasets.base_dataset import BaseDataset, DEFAULT_CACHE_LOCATION
from ludwig.datasets.mixins.download import GZipDownloadMixin
from ludwig.datasets.mixins.load import CSVLoadMixin

NUM_LABELS = 10


def load(cache_dir=DEFAULT_CACHE_LOCATION, split=False):
    dataset = Mnist(cache_dir=cache_dir)
    return dataset.load(split=split)


class Mnist(CSVLoadMixin, GZipDownloadMixin, BaseDataset):
    """The Mnist dataset.

    This pulls in an array of mixins for different types of functionality
    which belongs in the workflow for ingesting and transforming training data into a destination
    dataframe that can fit into Ludwig's training API.
    """

    config: dict
    raw_temp_path: str
    raw_dataset_path: str
    processed_temp_path: str
    processed_dataset_path: str

    def __init__(self, cache_dir=DEFAULT_CACHE_LOCATION):
        super().__init__(dataset_name="mnist", cache_dir=cache_dir)

    def process_downloaded_dataset(self):
        """Read the training and test directories and write out
        a csv containing the training path and the label.
        """
        os.makedirs(self.processed_temp_path, exist_ok=True)
        for dataset in ["training", "testing"]:
            print(f'>>> create ludwig formatted {dataset} data')
            labels, data = self.read_source_dataset(dataset,
                                                    self.raw_dataset_path)
            self.write_output_dataset(labels, data,
                                      os.path.join(self.processed_temp_path,
                                                   dataset))
        self.output_training_and_test_data()
        os.rename(self.processed_temp_path, self.processed_dataset_path)
        print('>>> completed data preparation')

    def read_source_dataset(self, dataset="training", path="."):
        """Create a directory for training and test and extract all the images
        and labels to this destination.
        :args:
            dataset (str) : the label for the dataset
            path (str): the raw dataset path
        :returns:
            A tuple of the label for the image, the file array, the size and rows and columns for the image"""
        if dataset == "training":
            fname_img = os.path.join(path, 'train-images-idx3-ubyte')
            fname_lbl = os.path.join(path, 'train-labels-idx1-ubyte')
        elif dataset == "testing":
            fname_img = os.path.join(path, 't10k-images-idx3-ubyte')
            fname_lbl = os.path.join(path, 't10k-labels-idx1-ubyte')
        else:
            raise ValueError("dataset must be 'testing' or 'training'")

        with open(fname_lbl, 'rb') as flbl:
            struct.unpack(">II", flbl.read(8))
            lbl = np.frombuffer(flbl.read(), dtype=np.uint8)

        with open(fname_img, 'rb') as fimg:
            magic_nr, size, rows, cols = struct.unpack(">IIII", fimg.read(16))
            img = np.frombuffer(fimg.read(), dtype=np.uint8)
            img = img.reshape((size, rows, cols))

        return lbl, img

    def write_output_dataset(self, labels, data, output_dir):
        """Create output directories where we write out the images.
        :args:
            labels (str) : the labels for the image
            data (np.array) : the binary array corresponding to the image
            output_dir (str) : the output directory that we need to write to
            path (str): the raw dataset path
        :returns:
            A tuple of the label for the image, the file array, the size and rows and columns for the image"""
        # create child image output directories
        output_dirs = [
            os.path.join(output_dir, str(i))
            for i in range(NUM_LABELS)
        ]

        for output_dir in output_dirs:
            os.makedirs(output_dir, exist_ok=True)

        def write_processed_image(t):
            i, label = t
            output_filename = os.path.join(output_dirs[label], str(i) + ".png")
            imsave(output_filename, data[i])

        # write out image data
        tasks = list(enumerate(labels))
        pool = ThreadPool(NUM_LABELS)
        pool.map(write_processed_image, tasks)
        pool.close()
        pool.join()

    def output_training_and_test_data(self):
        """The final method where we create a training and test file by iterating through
        all the images and labels previously created.
        """
        with open(
                os.path.join(self.processed_temp_path, self.csv_filename),
                'w'
        ) as output_file:
            output_file.write('image_path,label,split\n')
            for name in ["training", "testing"]:
                split = 0 if name == 'training' else 2
                for i in range(NUM_LABELS):
                    img_path = os.path.join(
                        self.processed_temp_path,
                        '{}/{}'.format(name, i)
                    )
                    final_img_path = os.path.join(
                        self.processed_dataset_path,
                        '{}/{}'.format(name, i)
                    )
                    for file in os.listdir(img_path):
                        if file.endswith(".png"):
                            output_file.write('{},{},{}\n'.format(
                                os.path.join(final_img_path, file),
                                str(i),
                                split
                            ))

    @property
    def download_url(self):
        return self.config["download_url"]



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os.path

import pytest

import mlflow
import ray
from mlflow.tracking import MlflowClient

from ludwig.contribs import MlflowCallback
from ludwig.hyperopt.execution import get_build_hyperopt_executor
from ludwig.hyperopt.results import RayTuneResults
from ludwig.hyperopt.run import hyperopt
from ludwig.hyperopt.sampling import (get_build_hyperopt_sampler)
from ludwig.hyperopt.run import update_hyperopt_params_with_defaults
from ludwig.utils.defaults import merge_with_defaults, ACCURACY
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import spawn
from tests.integration_tests.utils import text_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

HYPEROPT_CONFIG = {
    "parameters": {
        "training.learning_rate": {
            "space": "loguniform",
            "lower": 0.001,
            "upper": 0.1,
        },
        "combiner.num_fc_layers": {
            "space": "randint",
            "lower": 2,
            "upper": 6
        },
        "utterance.cell_type": {
            "space": "grid_search",
            "values": ["rnn", "gru"]
        },
        "utterance.bidirectional": {
            "space": "choice",
            "categories": [True, False]
        },
        "utterance.fc_layers": {
            "space": "choice",
            "categories": [
                [{"fc_size": 512}, {"fc_size": 256}],
                [{"fc_size": 512}],
                [{"fc_size": 256}],
            ]
        }
    },
    "goal": "minimize"
}


SAMPLERS = [
    {"type": "ray"},
    {"type": "ray", "num_samples": 2},
    {
        "type": "ray",
        "search_alg": {
            "type": "bohb"
        },
        "scheduler": {
            "type": "hb_bohb",
            "time_attr": "training_iteration",
            "reduction_factor": 4,
        },
        "num_samples": 3
    },
]

EXECUTORS = [
    {"type": "ray"},
]


def _get_config(sampler, executor):
    input_features = [
        text_feature(name="utterance", cell_type="lstm", reduce_output="sum"),
        category_feature(vocab_size=2, reduce_input="sum")]

    output_features = [category_feature(vocab_size=2, reduce_input="sum")]

    return {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        "training": {"epochs": 2, "learning_rate": 0.001},
        "hyperopt": {
            **HYPEROPT_CONFIG,
            "executor": executor,
            "sampler": sampler,
        },
    }


@pytest.fixture
def ray_start_4_cpus():
    address_info = ray.init(num_cpus=4)
    try:
        yield address_info
    finally:
        ray.shutdown()


@spawn
def run_hyperopt_executor(
    sampler, executor, csv_filename,
    validate_output_feature=False,
    validation_metric=None,
):
    config = _get_config(sampler, executor)
    rel_path = generate_data(
        config['input_features'],
        config['output_features'],
        csv_filename
    )

    config = merge_with_defaults(config)

    hyperopt_config = config["hyperopt"]

    if validate_output_feature:
        hyperopt_config['output_feature'] = config['output_features'][0]['name']
    if validation_metric:
        hyperopt_config['validation_metric'] = validation_metric

    update_hyperopt_params_with_defaults(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    if sampler.get("search_alg", {}).get("type", "") == 'bohb':
        # bohb does not support grid_search search space
        del parameters['utterance.cell_type']

    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]

    hyperopt_sampler = get_build_hyperopt_sampler(
        sampler["type"])(goal, parameters, **sampler)

    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        hyperopt_sampler, output_feature, metric, split, **executor)

    hyperopt_executor.execute(
        config,
        dataset=rel_path,
        backend='local',
    )


@pytest.mark.distributed
@pytest.mark.parametrize('sampler', SAMPLERS)
@pytest.mark.parametrize('executor', EXECUTORS)
def test_hyperopt_executor(sampler, executor, csv_filename, ray_start_4_cpus):
    run_hyperopt_executor(sampler, executor, csv_filename)


@pytest.mark.distributed
def test_hyperopt_executor_with_metric(csv_filename, ray_start_4_cpus):
    run_hyperopt_executor({"type": "ray", "num_samples": 2},
                          {"type": "ray"},
                          csv_filename,
                          validate_output_feature=True,
                          validation_metric=ACCURACY)


@pytest.mark.distributed
def test_hyperopt_run_hyperopt(csv_filename, ray_start_4_cpus):
    input_features = [
        text_feature(name="utterance", cell_type="lstm", reduce_output="sum"),
        category_feature(vocab_size=2, reduce_input="sum")]

    output_features = [category_feature(vocab_size=2, reduce_input="sum")]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        "training": {"epochs": 2, "learning_rate": 0.001}
    }

    output_feature_name = output_features[0]['name']

    hyperopt_configs = {
        "parameters": {
            "training.learning_rate": {
                "space": "loguniform",
                "lower": 0.001,
                "upper": 0.1,
            },
            output_feature_name + ".fc_size": {
                "space": "randint",
                "lower": 32,
                "upper": 256
            },
            output_feature_name + ".num_fc_layers": {
                "space": "randint",
                "lower": 2,
                "upper": 6
            }
        },
        "goal": "minimize",
        'output_feature': output_feature_name,
        'validation_metrics': 'loss',
        'executor': {'type': 'ray'},
        'sampler': {'type': 'ray', 'num_samples': 2}
    }

    # add hyperopt parameter space to the config
    config['hyperopt'] = hyperopt_configs
    run_hyperopt(config, rel_path)


@pytest.mark.distributed
def test_hyperopt_ray_mlflow(csv_filename, ray_start_4_cpus, tmpdir):
    mlflow_uri = f'file://{tmpdir}/mlruns'
    mlflow.set_tracking_uri(mlflow_uri)
    client = MlflowClient(tracking_uri=mlflow_uri)

    num_samples = 2
    config = _get_config(
        {"type": "ray", "num_samples": num_samples},
        {"type": "ray"}
    )

    rel_path = generate_data(
        config['input_features'],
        config['output_features'],
        csv_filename
    )

    exp_name = 'mlflow_test'
    run_hyperopt(config, rel_path,
                 experiment_name=exp_name,
                 callbacks=[MlflowCallback(mlflow_uri)])

    experiment = client.get_experiment_by_name(exp_name)
    assert experiment is not None

    runs = client.search_runs([experiment.experiment_id])
    assert len(runs) > 0

    for run in runs:
        artifacts = [f.path for f in client.list_artifacts(run.info.run_id, "")]
        assert 'config.yaml' in artifacts
        assert 'model' in artifacts


@spawn
def run_hyperopt(
        config, rel_path,
        experiment_name='ray_hyperopt',
        callbacks=None,
):
    hyperopt_results = hyperopt(
        config,
        dataset=rel_path,
        output_directory='results_hyperopt',
        experiment_name=experiment_name,
        callbacks=callbacks,
    )

    # check for return results
    assert isinstance(hyperopt_results, RayTuneResults)

    # check for existence of the hyperopt statistics file
    assert os.path.isfile(
        os.path.join('results_hyperopt', 'hyperopt_statistics.json')
    )



import json
import logging
import os


from ludwig.hyperopt.results import HyperoptResults
from ludwig.utils.data_utils import save_json
from ludwig.utils.print_utils import print_boxed

logger = logging.getLogger(__name__)


def print_hyperopt_results(hyperopt_results: HyperoptResults):
    print_boxed('HYPEROPT RESULTS', print_fun=logger.info)
    for trial_results in hyperopt_results.ordered_trials:
        logger.info('score: {:.6f} | parameters: {}'.format(
            trial_results.metric_score, trial_results.parameters
        ))
    logger.info("")


def save_hyperopt_stats(hyperopt_stats, hyperopt_dir_name):
    hyperopt_stats_fn = os.path.join(
        hyperopt_dir_name,
        'hyperopt_statistics.json'
    )
    save_json(hyperopt_stats_fn, hyperopt_stats)


def load_json_value(v):
    try:
        return json.loads(v)
    except:
        return v


def load_json_values(d):
    return {
        k: load_json_value(v)
        for k, v in d.items()
    }



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.datasets.base_dataset import BaseDataset, DEFAULT_CACHE_LOCATION
from ludwig.datasets.mixins.download import  TarDownloadMixin
from ludwig.datasets.mixins.load import CSVLoadMixin
from ludwig.datasets.mixins.process import *

def load(cache_dir=DEFAULT_CACHE_LOCATION, split=True):
    dataset = YelpReviews(cache_dir=cache_dir)
    return dataset.load(split=split)

class YelpReviews(TarDownloadMixin, MultifileJoinProcessMixin,
                 CSVLoadMixin, BaseDataset):
    """
        The Yelp Reviews dataset
        Details:
            1,569,264 samples from the Yelp Dataset Challenge 2015. \
            This full dataset has 130,000 training samples and 10,000 \
            testing samples in each star. 
        Dataset source: 
            Character-level Convolutional Networks for Text Classification
            Xiang Zhang et al., 2015
    """
    def __init__(self, cache_dir=DEFAULT_CACHE_LOCATION):
        super().__init__(dataset_name="yelp_reviews", cache_dir=cache_dir)

    def process_downloaded_dataset(self):
        super().process_downloaded_dataset(header=None)
        processed_df = pd.read_csv(os.path.join(self.processed_dataset_path,
                                                self.csv_filename))
        processed_df.columns = ['label', 'text', 'split']
        processed_df.to_csv(
            os.path.join(self.processed_dataset_path, self.csv_filename),
            index=False
        )


    



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from ludwig.backend.base import Backend, LocalBackend
from ludwig.utils.horovod_utils import has_horovodrun

try:
    import ray as _ray
except:
    _ray = None


LOCAL_BACKEND = LocalBackend()

LOCAL = 'local'
DASK = 'dask'
HOROVOD = 'horovod'
RAY = 'ray'

ALL_BACKENDS = [LOCAL, DASK, HOROVOD, RAY]


def _has_ray():
    if _ray is None:
        return False

    try:
        _ray.init('auto', ignore_reinit_error=True)
        return True
    except:
        return False


def get_local_backend(**kwargs):
    return LocalBackend(**kwargs)


def create_dask_backend(**kwargs):
    from ludwig.backend.dask import DaskBackend
    return DaskBackend(**kwargs)


def create_horovod_backend(**kwargs):
    from ludwig.backend.horovod import HorovodBackend
    return HorovodBackend(**kwargs)


def create_ray_backend(**kwargs):
    from ludwig.backend.ray import RayBackend
    return RayBackend(**kwargs)


backend_registry = {
    LOCAL: get_local_backend,
    DASK: create_dask_backend,
    HOROVOD: create_horovod_backend,
    RAY: create_ray_backend,
    None: get_local_backend,
}


def create_backend(type, **kwargs):
    if isinstance(type, Backend):
        return type

    if type is None and _has_ray():
        type = RAY
    elif type is None and has_horovodrun():
        type = HOROVOD

    return backend_registry[type](**kwargs)


def initialize_backend(backend):
    if isinstance(backend, dict):
        backend = create_backend(**backend)
    else:
        backend = create_backend(backend)
    backend.initialize()
    return backend



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import itertools
import os
import shutil
import tempfile

import numpy as np
import pandas as pd

from ludwig.api import LudwigModel
from ludwig.constants import BINARY, SEQUENCE, TEXT, SET
from ludwig.utils.neuropod_utils import export_neuropod
from ludwig.utils.strings_utils import str2bool
from tests.integration_tests.utils import category_feature, binary_feature, \
    numerical_feature, text_feature, set_feature, vector_feature, \
    image_feature, \
    audio_feature, timeseries_feature, date_feature, h3_feature, bag_feature, LocalTestBackend
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import sequence_feature


def test_neuropod(csv_filename):
    #######
    # Setup
    #######
    with tempfile.TemporaryDirectory() as tmpdir:
        dir_path = tmpdir
        data_csv_path = os.path.join(tmpdir, csv_filename)
        image_dest_folder = os.path.join(tmpdir, 'generated_images')
        audio_dest_folder = os.path.join(tmpdir, 'generated_audio')

        input_features = [
            binary_feature(),
            numerical_feature(),
            category_feature(vocab_size=3),
            sequence_feature(vocab_size=3),
            text_feature(vocab_size=3),
            vector_feature(),
            image_feature(image_dest_folder),
            audio_feature(audio_dest_folder),
            timeseries_feature(),
            date_feature(),
            h3_feature(),
            set_feature(vocab_size=3),
            bag_feature(vocab_size=3),
        ]

        output_features = [
            binary_feature(),
            numerical_feature(),
            category_feature(vocab_size=3),
            sequence_feature(vocab_size=3),
            text_feature(vocab_size=3),
            set_feature(vocab_size=3),
            vector_feature()
        ]

        # Generate test data
        data_csv_path = generate_data(input_features, output_features,
                                      data_csv_path)

        #############
        # Train model
        #############
        config = {
            'input_features': input_features,
            'output_features': output_features,
            'training': {'epochs': 2}
        }
        ludwig_model = LudwigModel(config, backend=LocalTestBackend())
        ludwig_model.train(
            dataset=data_csv_path,
            skip_save_training_description=True,
            skip_save_training_statistics=True,
            skip_save_progress=True,
            skip_save_log=True,
            skip_save_processed_input=True,
            output_directory=dir_path
        )

        data_df = pd.read_csv(data_csv_path)
        original_predictions_df, _ = ludwig_model.predict(dataset=data_df)

        ###################
        # save Ludwig model
        ###################
        ludwigmodel_path = os.path.join(dir_path, 'ludwigmodel')
        shutil.rmtree(ludwigmodel_path, ignore_errors=True)
        ludwig_model.save(ludwigmodel_path)

        ################
        # build neuropod
        ################
        neuropod_path = os.path.join(dir_path, 'neuropod')
        shutil.rmtree(neuropod_path, ignore_errors=True)
        export_neuropod(
            ludwigmodel_path, neuropod_path=neuropod_path, entrypoint='get_test_model'
        )

        ########################
        # predict using neuropod
        ########################
        if_dict = {
            input_feature['name']: np.expand_dims(np.array(
                [str(x) for x in data_df[input_feature['name']].tolist()],
                dtype='str'
            ), 1)
            for input_feature in input_features
        }

        from neuropod.loader import load_neuropod
        neuropod_model = load_neuropod(neuropod_path, _always_use_native=False)
        preds = neuropod_model.infer(if_dict)

        for key in preds:
            preds[key] = np.squeeze(preds[key])

        #########
        # cleanup
        #########
        # Delete the temporary data created
        for path in [ludwigmodel_path, neuropod_path,
                     image_dest_folder, audio_dest_folder]:
            if os.path.exists(path):
                if os.path.isfile(path):
                    os.remove(path)
                else:
                    shutil.rmtree(path, ignore_errors=True)

        ########
        # checks
        ########
        for output_feature in output_features:
            output_feature_name = output_feature['name']
            output_feature_type = output_feature['type']

            if (output_feature_name + "_predictions" in preds and
                    output_feature_name + "_predictions" in original_predictions_df):
                neuropod_pred = preds[
                    output_feature_name + "_predictions"].tolist()
                if output_feature_type == BINARY:
                    neuropod_pred = [ str2bool(x) for x in neuropod_pred]
                if output_feature_type in {SEQUENCE, TEXT, SET}:
                    neuropod_pred = [ x.split() for x in neuropod_pred]

                original_pred = original_predictions_df[
                    output_feature_name + "_predictions"].tolist()

                assert neuropod_pred == original_pred

            if (output_feature_name + "_probability" in preds and
                    output_feature_name + "_probability" in original_predictions_df):
                neuropod_prob = preds[
                    output_feature_name + "_probability"].tolist()
                if output_feature_type in {SEQUENCE, TEXT, SET}:
                    neuropod_prob = [ [float(n) for n in x.split()] for x in neuropod_prob]
                if any(isinstance(el, list) for el in neuropod_prob):
                    neuropod_prob = np.array(list(
                        itertools.zip_longest(*neuropod_prob, fillvalue=0)
                    )).T

                original_prob = original_predictions_df[
                    output_feature_name + "_probability"].tolist()
                if any(isinstance(el, list) for el in original_prob):
                    original_prob = np.array(list(
                        itertools.zip_longest(*original_prob, fillvalue=0)
                    )).T

                assert np.allclose(neuropod_prob, original_prob)

            if (output_feature_name + "_probabilities" in preds and
                    output_feature_name + "_probabilities" in original_predictions_df):
                neuropod_prob = preds[
                    output_feature_name + "_probabilities"].tolist()

                original_prob = original_predictions_df[
                    output_feature_name + "_probabilities"].tolist()

                assert np.allclose(neuropod_prob, original_prob)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

from tensorflow.keras.layers import Layer

from ludwig.modules.fully_connected_modules import FCStack

logger = logging.getLogger(__name__)


class PassthroughEncoder(Layer):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__()
        logger.debug(' {}'.format(self.name))

    def call(self, inputs, training=None, mask=None):
        """
            :param inputs: The inputs fed into the encoder.
                   Shape: [batch x 1], type tf.float32
        """
        return {'encoder_output': inputs}


class DenseEncoder(Layer):

    def __init__(
            self,
            layers=None,
            num_layers=1,
            fc_size=256,
            use_bias=True,
            weights_initializer='glorot_uniform',
            bias_initializer='zeros',
            weights_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            # weights_constraint=None,
            # bias_constraint=None,
            norm=None,
            norm_params=None,
            activation='relu',
            dropout=0,
            **kwargs
    ):
        super().__init__()
        logger.debug(' {}'.format(self.name))

        logger.debug('  FCStack')
        self.fc_stack = FCStack(
            layers=layers,
            num_layers=num_layers,
            default_fc_size=fc_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_weights_regularizer=weights_regularizer,
            default_bias_regularizer=bias_regularizer,
            default_activity_regularizer=activity_regularizer,
            # default_weights_constraint=weights_constraint,
            # default_bias_constraint=bias_constraint,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def call(self, inputs, training=None, mask=None):
        """
            :param inputs: The inputs fed into the encoder.
                   Shape: [batch x 1], type tf.float32
        """
        return {'encoder_output': self.fc_stack(inputs)}



import os
import shutil
import tempfile
from collections import namedtuple

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf

from ludwig.constants import PROC_COLUMN, NAME
from ludwig.data.dataset_synthesizer import build_synthetic_dataset
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.features.feature_utils import SEQUENCE_TYPES, compute_feature_hash
from ludwig.models.ecd import build_single_input, build_single_output
from tests.integration_tests.utils import category_feature, LocalTestBackend
from tests.integration_tests.utils import date_feature
from tests.integration_tests.utils import image_feature
from tests.integration_tests.utils import numerical_feature
from tests.integration_tests.utils import sequence_feature
from tests.integration_tests.utils import set_feature

BATCH_SIZE = 32
HIDDEN_SIZE = 128
SEQ_SIZE = 10
RANDOM_SEED = 42
IMAGE_DIR = tempfile.mkdtemp()

# SyntheticData namedtuple structure:
# batch_size: Number of records to generate for a batch
# feature_generator: Ludwig synthetic generator class
# feature_generator_args: tuple of required positional arguments
# feature_generator_kwargs: dictionary of required keyword arguments
SyntheticData = namedtuple(
    'SyntheticData',
    'batch_size feature_generator feature_generator_args feature_generator_kwargs'
)

# TestCase namedtuple structure:
# syn_data: SyntheticData namedtuple of data to create
# XCoder_other_parms: dictionary for required encoder/decoder parameters
# regularizer_parm_names: list of regularizer keyword parameter names
TestCase = namedtuple('TestCase',
                      'syn_data XCoder_other_parms regularizer_parm_names')


#
# Regularization Encoder Tests
#
@pytest.mark.parametrize(
    'test_case',
    [
        # DenseEncoder
        TestCase(
            SyntheticData(BATCH_SIZE, numerical_feature, (), {}),
            {'num_layers': 2, 'encoder': 'dense',
             'preprocessing': {'normalization': 'zscore'}},
            ['activity_regularizer', 'weights_regularizer', 'bias_regularizer']
        ),

        # Image Encoders
        TestCase(
            SyntheticData(BATCH_SIZE, image_feature, (IMAGE_DIR,), {}),
            {'encoder': 'stacked_cnn'},
            [
                'conv_activity_regularizer', 'conv_weights_regularizer',
                'conv_bias_regularizer',
                'fc_activity_regularizer', 'fc_weights_regularizer',
                'fc_bias_regularizer',
            ]
        ),
        TestCase(
            SyntheticData(BATCH_SIZE, image_feature, (IMAGE_DIR,), {}),
            {'encoder': 'resnet'},
            [
                'activity_regularizer', 'weights_regularizer',
                'bias_regularizer',
            ]
        ),

        # Categorical encoder
        TestCase(
            SyntheticData(BATCH_SIZE, category_feature, (), {}),
            {'representation': 'dense'},
            [
                'embedding_regularizer',
            ]
        ),

        # Date encoder
        TestCase(
            SyntheticData(BATCH_SIZE, date_feature, (), {}),
            {},
            [
                'activity_regularizer', 'weights_regularizer',
                'bias_regularizer',
            ]
        ),

        # ParallelCNN Encoder
        TestCase(
            SyntheticData(BATCH_SIZE, sequence_feature, (), {}),
            {'encoder': 'parallel_cnn', 'cell_type': 'gru'},
            [
                'activity_regularizer', 'weights_regularizer',
                'bias_regularizer',
            ]
        ),

        # Set Encoder
        TestCase(
            SyntheticData(BATCH_SIZE, set_feature, (), {}),
            {},
            [
                'activity_regularizer', 'weights_regularizer',
                'bias_regularizer',
            ]
        ),

    ]

)
def test_encoder(test_case):
    # set up required directories for images if needed
    shutil.rmtree(IMAGE_DIR, ignore_errors=True)
    os.mkdir(IMAGE_DIR)

    # reproducible synthetic data set
    np.random.seed(RANDOM_SEED)
    tf.random.set_seed(RANDOM_SEED)

    # create synthetic data for the test
    features = [
        test_case.syn_data.feature_generator(
            *test_case.syn_data.feature_generator_args,
            **test_case.syn_data.feature_generator_kwargs
        )
    ]
    name = features[0][NAME]
    proc_column = compute_feature_hash(features[0])
    features[0][PROC_COLUMN] = proc_column

    data_generator = build_synthetic_dataset(BATCH_SIZE, features)
    data_list = list(data_generator)
    raw_data = [x[0] for x in data_list[1:]]
    df = pd.DataFrame({data_list[0][0]: raw_data})

    # minimal config sufficient to create the input feature
    config = {'input_features': features, 'output_features': []}
    training_set, _, _, training_set_metadata = preprocess_for_training(
        config,
        training_set=df,
        skip_save_processed_input=True,
        random_seed=RANDOM_SEED,
        backend=LocalTestBackend(),
    )

    # run through each type of regularizer for the encoder
    regularizer_losses = []
    for regularizer in [None, 'l1', 'l2', 'l1_l2']:
        # start with clean slate and make reproducible
        tf.keras.backend.clear_session()
        np.random.seed(RANDOM_SEED)
        tf.random.set_seed(RANDOM_SEED)

        # setup kwarg for regularizer parms
        x_coder_kwargs = dict(
            zip(test_case.regularizer_parm_names,
                len(test_case.regularizer_parm_names) * [regularizer])
        )

        # combine other other keyword parameters
        x_coder_kwargs.update(test_case.XCoder_other_parms)
        features[0].update(x_coder_kwargs)

        # shim code to support sequence/sequence like features
        if features[0]['type'] in SEQUENCE_TYPES.union({'category', 'set'}):
            features[0]['vocab'] = training_set_metadata[name][
                'idx2str']
            training_set.dataset[proc_column] = \
                training_set.dataset[proc_column].astype(np.int32)

        input_def_obj = build_single_input(features[0], None)

        inputs = training_set.dataset[proc_column]
        # make sure we are at least rank 2 tensor
        if len(inputs.shape) == 1:
            inputs = inputs.reshape(-1, 1)

        # special handling for image feature
        if features[0]['type'] == 'image':
            inputs = tf.cast(inputs, tf.float32) / 255

        input_def_obj.encoder_obj(inputs)
        regularizer_loss = tf.reduce_sum(input_def_obj.encoder_obj.losses)
        regularizer_losses.append(regularizer_loss)

    # check loss regularization loss values
    # None should be zero
    assert regularizer_losses[0] == 0

    # l1, l2 and l1_l2 should be greater than zero
    assert np.all([t > 0.0 for t in regularizer_losses[1:]])

    # # using default setting l1 + l2 == l1_l2 losses
    assert np.isclose(
        regularizer_losses[1].numpy() + regularizer_losses[2].numpy(),
        regularizer_losses[3].numpy())

    # cleanup
    shutil.rmtree(IMAGE_DIR, ignore_errors=True)


#
# Regularization Decoder Tests
#
@pytest.mark.parametrize(
    'test_case',
    [
        # regressor decoder
        TestCase(
            SyntheticData(BATCH_SIZE, numerical_feature, (), {}),
            {
                'decoder': 'regressor',
                'loss': {'type': 'mean_squared_error'},
                'num_fc_layers': 5
            },
            ['activity_regularizer', 'weights_regularizer', 'bias_regularizer']
        ),

        # Tagger Decoder
        TestCase(
            SyntheticData(BATCH_SIZE, sequence_feature, (),
                          {'max_len': SEQ_SIZE}),
            {'decoder': 'tagger'},
            ['activity_regularizer', 'weights_regularizer', 'bias_regularizer']
        ),

        # Generator Decoder
        TestCase(
            SyntheticData(BATCH_SIZE, sequence_feature, (),
                          {'max_len': SEQ_SIZE}),
            {'decoder': 'generator', 'cell_type': 'gru'},
            ['activity_regularizer', 'weights_regularizer', 'bias_regularizer']
        ),

    ]

)
def test_decoder(test_case):
    # reproducible synthetic data set
    np.random.seed(RANDOM_SEED)
    tf.random.set_seed(RANDOM_SEED)

    # create synthetic data for the test
    features = [
        test_case.syn_data.feature_generator(
            *test_case.syn_data.feature_generator_args,
            **test_case.syn_data.feature_generator_kwargs
        )
    ]
    feature_name = features[0][NAME]
    proc_column = compute_feature_hash(features[0])
    features[0][PROC_COLUMN] = proc_column

    data_generator = build_synthetic_dataset(BATCH_SIZE, features)
    data_list = list(data_generator)
    raw_data = [x[0] for x in data_list[1:]]
    df = pd.DataFrame({data_list[0][0]: raw_data})

    # create synthetic combiner layer
    combiner_outputs_rank2 = {
        'combiner_output': tf.random.normal(
            [BATCH_SIZE, HIDDEN_SIZE],
            dtype=tf.float32
        )
    }

    combiner_outputs_rank3 = {
        'combiner_output': tf.random.normal(
            [BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE],
            dtype=tf.float32
        ),
        'encoder_output_state': tf.random.normal(
            [BATCH_SIZE, HIDDEN_SIZE],
            dtype=tf.float32
        ),
        'lengths': tf.convert_to_tensor(
            np.array(BATCH_SIZE * [SEQ_SIZE]),
            dtype=tf.int32
        )
    }

    # minimal config sufficient to create output feature
    config = {'input_features': [], 'output_features': features}
    training_set, _, _, training_set_metadata = preprocess_for_training(
        config,
        training_set=df,
        skip_save_processed_input=True,
        random_seed=RANDOM_SEED
    )

    # run through each type of regularizer
    regularizer_losses = []
    for regularizer in [None, 'l1', 'l2', 'l1_l2']:
        # start with clean slate and make reproducible
        tf.keras.backend.clear_session()
        np.random.seed(RANDOM_SEED)
        tf.random.set_seed(RANDOM_SEED)

        # setup kwarg for regularizer parms
        x_coder_kwargs = dict(
            zip(test_case.regularizer_parm_names,
                len(test_case.regularizer_parm_names) * [regularizer])
        )

        # combine other other keyword parameters
        x_coder_kwargs.update(test_case.XCoder_other_parms)

        features[0].update(x_coder_kwargs)
        if features[0]['type'] in SEQUENCE_TYPES:
            features[0]['num_classes'] = training_set_metadata[feature_name][
                                             'vocab_size'] + 1
            training_set.dataset[proc_column] = \
                training_set.dataset[proc_column].astype(np.int32)
            combiner_outputs = combiner_outputs_rank3
        else:
            combiner_outputs = combiner_outputs_rank2

        output_def_obj = build_single_output(features[0], None, None)

        targets = training_set.dataset[proc_column]
        if len(targets.shape) == 1:
            targets = targets.reshape(-1, 1)

        output_def_obj(
            (
                (combiner_outputs, None),
                targets
            ),
            training=True,
            mask=None
        )
        regularizer_loss = tf.reduce_sum(output_def_obj.decoder_obj.losses)
        regularizer_losses.append(regularizer_loss)

    # check loss regularization loss values
    # None should be zero
    assert regularizer_losses[0] == 0

    # l1, l2 and l1_l2 should be greater than zero
    assert np.all([t > 0.0 for t in regularizer_losses[1:]])

    # # using default setting l1 + l2 == l1_l2 losses
    assert np.isclose(
        regularizer_losses[1].numpy() + regularizer_losses[2].numpy(),
        regularizer_losses[3].numpy())



# coding=utf-8
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import time

try:
    import horovod.tensorflow

    _HVD = horovod.tensorflow
except (ModuleNotFoundError, ImportError):
    _HVD = None


def initialize_horovod():
    if not _HVD:
        raise ValueError("Horovod backend specified, "
                         "but cannot import `horovod.tensorflow`. "
                         "Install Horovod following the instructions at: "
                         "https://github.com/horovod/horovod")
    _HVD.init()
    return _HVD


def has_horovodrun():
    """Returns True if running with `horovodrun` using Gloo or OpenMPI."""
    return 'OMPI_COMM_WORLD_RANK' in os.environ or 'HOROVOD_RANK' in os.environ


def return_first(fn):
    """Wraps function so results are only returned by the first (coordinator) rank.

    The purpose of this function is to reduce network overhead.
    """
    def wrapped(*args, **kwargs):
        res = fn(*args, **kwargs)
        return res if _HVD.rank() == 0 else None
    return wrapped



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
import sys

import numpy as np
import tensorflow as tf

from ludwig.constants import *
from ludwig.encoders.sequence_encoders import StackedCNN, ParallelCNN, \
    StackedParallelCNN, StackedRNN, SequencePassthroughEncoder, StackedCNNRNN
from ludwig.features.sequence_feature import SequenceInputFeature
from ludwig.utils.audio_utils import calculate_mean, calculate_var
from ludwig.utils.audio_utils import get_fbank
from ludwig.utils.audio_utils import get_group_delay
from ludwig.utils.audio_utils import get_length_in_samp
from ludwig.utils.audio_utils import get_max_length_stft_based
from ludwig.utils.audio_utils import get_non_symmetric_length
from ludwig.utils.audio_utils import get_phase_stft_magnitude
from ludwig.utils.audio_utils import get_stft_magnitude
from ludwig.utils.data_utils import get_abs_path
from ludwig.utils.misc_utils import set_default_value
from ludwig.utils.misc_utils import set_default_values

logger = logging.getLogger(__name__)


class AudioFeatureMixin:
    type = AUDIO

    preprocessing_defaults = {
        'audio_file_length_limit_in_s': 7.5,
        'missing_value_strategy': BACKFILL,
        'in_memory': True,
        'padding_value': 0,
        'norm': None,
        'audio_feature': {
            TYPE: 'raw',
        }
    }

    preprocessing_schema = {
        'audio_file_length_limit_in_s': {'type': 'number', 'minimum': 0},
        'missing_value_strategy': {'type': 'string', 'enum': MISSING_VALUE_STRATEGY_OPTIONS},
        'in_memory': {'type': 'boolean'},
        'padding_value': {'type': 'number', 'minimum': 0},
        'norm': {'type': ['string', 'null'], 'enum': [None, 'per_file', 'global']},
        'audio_feature': {
            'type': 'object',
            'properties': {
                'type': {'type': 'string', 'enum': ['raw', 'stft', 'stft_phase', 'group_delay', 'fbank']},
                'window_length_in_s': {'type': 'number', 'minimum': 0},
                'window_shift_in_s': {'type': 'number', 'minimum': 0},
                'num_fft_points': {'type': 'number', 'minimum': 0},
                'window_type': {'type': 'string'},
                'num_filter_bands': {'type': 'number', 'minimum': 0},
            }
        }
    }

    @staticmethod
    def cast_column(column, backend):
        return column

    @staticmethod
    def get_feature_meta(column, preprocessing_parameters, backend):
        try:
            import soundfile
        except ImportError:
            logger.error(
                ' soundfile is not installed. '
                'In order to install all audio feature dependencies run '
                'pip install ludwig[audio]'
            )
            sys.exit(-1)

        audio_feature_dict = preprocessing_parameters['audio_feature']
        first_audio_file_path = column.head(1)[0]
        _, sampling_rate_in_hz = soundfile.read(first_audio_file_path)

        feature_dim = AudioFeatureMixin._get_feature_dim(audio_feature_dict,
                                                         sampling_rate_in_hz)
        audio_file_length_limit_in_s = preprocessing_parameters[
            'audio_file_length_limit_in_s']
        max_length = AudioFeatureMixin._get_max_length_feature(
            audio_feature_dict, sampling_rate_in_hz,
            audio_file_length_limit_in_s)
        return {
            'feature_dim': feature_dim,
            'sampling_rate_in_hz': sampling_rate_in_hz,
            'max_length': max_length,
            'reshape': (max_length, feature_dim)
        }

    @staticmethod
    def _get_feature_dim(audio_feature_dict, sampling_rate_in_hz):
        feature_type = audio_feature_dict[TYPE]

        if feature_type == 'raw':
            feature_dim = 1
        elif feature_type == 'stft_phase':
            feature_dim_symmetric = get_length_in_samp(
                audio_feature_dict['window_length_in_s'], sampling_rate_in_hz)
            feature_dim = 2 * get_non_symmetric_length(feature_dim_symmetric)
        elif feature_type in ['stft', 'group_delay']:
            feature_dim_symmetric = get_length_in_samp(
                audio_feature_dict['window_length_in_s'], sampling_rate_in_hz)
            feature_dim = get_non_symmetric_length(feature_dim_symmetric)
        elif feature_type == 'fbank':
            feature_dim = audio_feature_dict['num_filter_bands']
        else:
            raise ValueError('{} is not recognized.'.format(feature_type))

        return feature_dim

    @staticmethod
    def _process_in_memory(
            column,
            src_path,
            audio_feature_dict,
            feature_dim,
            max_length,
            padding_value,
            normalization_type,
            audio_file_length_limit_in_s,
            backend
    ):
        try:
            import soundfile
        except ImportError:
            logger.error(
                ' soundfile is not installed. '
                'In order to install all audio feature dependencies run '
                'pip install ludwig[audio]'
            )
            sys.exit(-1)

        def read_audio(path):
            filepath = get_abs_path(src_path, path)
            return soundfile.read(filepath)

        df_engine = backend.df_engine
        raw_audio = df_engine.map_objects(column, read_audio)
        processed_audio = df_engine.map_objects(
            raw_audio,
            lambda row: AudioFeatureMixin._transform_to_feature(
                audio=row[0],
                sampling_rate_in_hz=row[1],
                audio_feature_dict=audio_feature_dict,
                feature_dim=feature_dim,
                max_length=max_length,
                padding_value=padding_value,
                normalization_type=normalization_type
            )
        )

        audio_stats = df_engine.map_objects(
            raw_audio,
            lambda row: AudioFeatureMixin._get_stats(
                audio=row[0],
                sampling_rate_in_hz=row[1],
                max_length_in_s=audio_file_length_limit_in_s,
            )
        )

        def reduce(series):
            merged_stats = None
            for audio_stats in series:
                if merged_stats is None:
                    merged_stats = audio_stats.copy()
                else:
                    AudioFeatureMixin._merge_stats(merged_stats, audio_stats)
            return merged_stats

        merged_stats = df_engine.reduce_objects(audio_stats, reduce)
        merged_stats['mean'] = calculate_mean(merged_stats['sum'], merged_stats['count'])
        merged_stats['var'] = calculate_var(merged_stats['sum'], merged_stats['sum2'], merged_stats['count'])
        return processed_audio, merged_stats

    @staticmethod
    def _transform_to_feature(
            audio,
            sampling_rate_in_hz,
            audio_feature_dict,
            feature_dim,
            max_length,
            padding_value,
            normalization_type
    ):
        feature_type = audio_feature_dict[TYPE]
        if feature_type == 'raw':
            audio_feature = np.expand_dims(audio, axis=-1)
        elif feature_type in ['stft', 'stft_phase', 'group_delay', 'fbank']:
            audio_feature = np.transpose(
                AudioFeatureMixin._get_2D_feature(audio, feature_type,
                                                  audio_feature_dict,
                                                  sampling_rate_in_hz))
        else:
            raise ValueError('{} is not recognized.'.format(feature_type))

        if normalization_type == 'per_file':
            mean = np.mean(audio_feature, axis=0)
            std = np.std(audio_feature, axis=0)
            audio_feature = np.divide((audio_feature - mean),
                                      std + 1.0e-10)
        elif normalization_type == 'global':
            raise ValueError('not implemented yet')

        feature_length = audio_feature.shape[0]
        broadcast_feature_length = min(feature_length, max_length)
        audio_feature_padded = np.full((max_length, feature_dim),
                                       padding_value,
                                       dtype=np.float32)
        audio_feature_padded[:broadcast_feature_length, :] = audio_feature[
                                                             :max_length, :]

        return audio_feature_padded

    @staticmethod
    def _get_stats(audio, sampling_rate_in_hz, max_length_in_s):
        audio_length_in_s = audio.shape[-1] / float(sampling_rate_in_hz)
        return {
            'count': 1,
            'sum': audio_length_in_s,
            'sum2': audio_length_in_s * audio_length_in_s,
            'min': audio_length_in_s,
            'max': audio_length_in_s,
            'cropped': 1 if audio_length_in_s > max_length_in_s else 0
        }

    @staticmethod
    def _merge_stats(merged_stats, audio_stats):
        merged_stats['count'] += audio_stats['count']
        merged_stats['sum'] += audio_stats['sum']
        merged_stats['sum2'] += audio_stats['sum2']
        merged_stats['min'] = min(merged_stats['min'], audio_stats['min'])
        merged_stats['max'] = max(merged_stats['max'], audio_stats['max'])
        merged_stats['cropped'] += audio_stats['cropped']

    @staticmethod
    def _get_2D_feature(audio, feature_type, audio_feature_dict,
                        sampling_rate_in_hz):
        window_length_in_s = audio_feature_dict['window_length_in_s']
        window_shift_in_s = audio_feature_dict['window_shift_in_s']
        window_length_in_samp = get_length_in_samp(window_length_in_s,
                                                   sampling_rate_in_hz)

        if 'num_fft_points' in audio_feature_dict:
            num_fft_points = audio_feature_dict['num_fft_points']
            if num_fft_points < window_length_in_samp:
                raise ValueError(
                    'num_fft_points: {} < window length in '
                    'samples: {} (corresponds to window length'
                    ' in s: {}'.format(num_fft_points, window_length_in_s,
                                       window_length_in_samp))
        else:
            num_fft_points = window_length_in_samp

        if 'window_type' in audio_feature_dict:
            window_type = audio_feature_dict['window_type']
        else:
            window_type = 'hamming'

        if feature_type == 'stft_phase':
            return get_phase_stft_magnitude(audio, sampling_rate_in_hz,
                                            window_length_in_s,
                                            window_shift_in_s, num_fft_points,
                                            window_type)
        if feature_type == 'stft':
            return get_stft_magnitude(audio, sampling_rate_in_hz,
                                      window_length_in_s, window_shift_in_s,
                                      num_fft_points, window_type)
        if feature_type == 'group_delay':
            return get_group_delay(audio, sampling_rate_in_hz,
                                   window_length_in_s, window_shift_in_s,
                                   num_fft_points, window_type)
        if feature_type == 'fbank':
            num_filter_bands = audio_feature_dict['num_filter_bands']
            return get_fbank(audio, sampling_rate_in_hz,
                             window_length_in_s, window_shift_in_s,
                             num_fft_points, window_type, num_filter_bands)

    @staticmethod
    def add_feature_data(
            feature,
            input_df,
            proc_df,
            metadata,
            preprocessing_parameters,
            backend,
            skip_save_processed_input
    ):
        set_default_value(
            feature['preprocessing'],
            'in_memory',
            preprocessing_parameters['in_memory']
        )

        if 'audio_feature' not in preprocessing_parameters:
            raise ValueError(
                'audio_feature dictionary has to be present in preprocessing '
                'for audio.')
        if TYPE not in preprocessing_parameters['audio_feature']:
            raise ValueError(
                'type has to be present in audio_feature dictionary '
                'for audio.')

        name = feature[NAME]
        column = feature[COLUMN]
        proc_column = feature[PROC_COLUMN]

        src_path = None
        # this is not super nice, but works both and DFs and lists
        first_path = '.'
        for first_path in input_df[column]:
            break
        if SRC in metadata:
            src_path = os.path.dirname(os.path.abspath(metadata.get(SRC)))
        if src_path is None and not os.path.isabs(first_path):
            raise ValueError('Audio file paths must be absolute')

        num_audio_utterances = len(input_df[feature[COLUMN]])
        padding_value = preprocessing_parameters['padding_value']
        normalization_type = preprocessing_parameters['norm']

        feature_dim = metadata[name]['feature_dim']
        max_length = metadata[name]['max_length']
        audio_feature_dict = preprocessing_parameters['audio_feature']
        audio_file_length_limit_in_s = preprocessing_parameters[
            'audio_file_length_limit_in_s']

        if num_audio_utterances == 0:
            raise ValueError(
                'There are no audio files in the dataset provided.')

        if feature[PREPROCESSING]['in_memory']:
            audio_features, audio_stats = AudioFeatureMixin._process_in_memory(
                input_df[feature[NAME]],
                src_path,
                audio_feature_dict,
                feature_dim,
                max_length,
                padding_value,
                normalization_type,
                audio_file_length_limit_in_s,
                backend
            )
            proc_df[proc_column] = audio_features

            audio_stats['std'] = np.sqrt(
                audio_stats['var'] / float(audio_stats['count']))
            print_statistics = (
                "{} audio files loaded.\n"
                "Statistics of audio file lengths:\n"
                "- mean: {:.4f}\n"
                "- std: {:.4f}\n"
                "- max: {:.4f}\n"
                "- min: {:.4f}\n"
                "- cropped audio_files: {}\n"
                "Max length was given as {}s"
            ).format(
                audio_stats['count'], audio_stats['mean'],
                audio_stats['std'], audio_stats['max'],
                audio_stats['min'], audio_stats['cropped'],
                audio_file_length_limit_in_s)
            logger.debug(print_statistics)
        else:
            backend.check_lazy_load_supported(feature)

        return proc_df

    @staticmethod
    def _get_max_length_feature(
            audio_feature_dict,
            sampling_rate_in_hz,
            audio_length_limit_in_s
    ):
        feature_type = audio_feature_dict[TYPE]
        audio_length_limit_in_samp = (
                audio_length_limit_in_s * sampling_rate_in_hz
        )

        if not audio_length_limit_in_samp.is_integer():
            raise ValueError(
                'Audio_file_length_limit has to be chosen '
                'so that {} (in s) * {} (sampling rate in Hz) '
                'is an integer.'.format(
                    audio_length_limit_in_s, sampling_rate_in_hz))
        audio_length_limit_in_samp = int(audio_length_limit_in_samp)

        if feature_type == 'raw':
            return audio_length_limit_in_samp
        elif feature_type in ['stft', 'stft_phase', 'group_delay', 'fbank']:
            window_length_in_s = audio_feature_dict['window_length_in_s']
            window_shift_in_s = audio_feature_dict['window_shift_in_s']
            return get_max_length_stft_based(audio_length_limit_in_samp,
                                             window_length_in_s,
                                             window_shift_in_s,
                                             sampling_rate_in_hz)
        else:
            raise ValueError('{} is not recognized.'.format(feature_type))


class AudioInputFeature(AudioFeatureMixin, SequenceInputFeature):
    encoder = 'parallel_cnn'
    max_sequence_length = None
    embedding_size = None

    def __init__(self, feature, encoder_obj=None):
        super().__init__(feature, encoder_obj=encoder_obj)
        if not self.embedding_size:
            raise ValueError(
                'embedding_size has to be defined - '
                'check "update_config_with_metadata()"')
        if not self.max_sequence_length:
            raise ValueError(
                'max_sequence_length has to be defined - '
                'check "update_config_with_metadata()"')

    def call(self, inputs, training=None, mask=None):
        assert isinstance(inputs, tf.Tensor)
        assert inputs.dtype == tf.float32
        assert len(inputs.shape) == 3

        encoder_output = self.encoder_obj(
            inputs, training=training, mask=mask
        )

        return encoder_output

    @classmethod
    def get_input_dtype(cls):
        return tf.float32

    def get_input_shape(self):
        return self.max_sequence_length, self.embedding_size

    @staticmethod
    def update_config_with_metadata(
            input_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        input_feature['max_sequence_length'] = feature_metadata['max_length']
        input_feature['embedding_size'] = feature_metadata['feature_dim']
        input_feature['should_embed'] = False

    @staticmethod
    def populate_defaults(input_feature):
        set_default_values(
            input_feature,
            {
                TIED: None,
                'preprocessing': {}
            }
        )

    encoder_registry = {
        'stacked_cnn': StackedCNN,
        'parallel_cnn': ParallelCNN,
        'stacked_parallel_cnn': StackedParallelCNN,
        'rnn': StackedRNN,
        'cnnrnn': StackedCNNRNN,
        'passthrough': SequencePassthroughEncoder,
        'null': SequencePassthroughEncoder,
        'none': SequencePassthroughEncoder,
        'None': SequencePassthroughEncoder,
        None: SequencePassthroughEncoder
    }



import os

from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
from dask.delayed import Delayed
from dask.utils import apply

from ludwig.data.dataframe.pandas import pandas_df_to_tfrecords, write_meta
from ludwig.data.dataset.tfrecord import get_part_filename, get_compression_ext
from ludwig.utils.fs_utils import makedirs


def dask_to_tfrecords(
        df,
        folder,
        compression_type="GZIP",
        compression_level=9):
    """Store Dask.dataframe to TFRecord files."""
    makedirs(folder, exist_ok=True)
    compression_ext = get_compression_ext(compression_type)
    filenames = [get_part_filename(i, compression_ext) for i in range(df.npartitions)]

    # Also write a meta data file
    write_meta(df, folder, compression_type)

    dsk = {}
    name = "to-tfrecord-" + tokenize(df, folder)
    part_tasks = []
    kwargs = {}

    for d, filename in enumerate(filenames):
        dsk[(name, d)] = (
            apply,
            pandas_df_to_tfrecords,
            [
                (df._name, d),
                os.path.join(folder, filename),
                compression_type,
                compression_level
            ],
            kwargs
        )
        part_tasks.append((name, d))

    dsk[name] = (lambda x: None, part_tasks)

    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[df])
    out = Delayed(name, graph)
    out = out.compute()
    return out



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import ABC

from ludwig.encoders.base import Encoder
from ludwig.encoders.generic_encoders import PassthroughEncoder
from ludwig.modules.embedding_modules import Embed
from ludwig.utils.registry import Registry, register, DEFAULT_KEYS

logger = logging.getLogger(__name__)


ENCODER_REGISTRY = Registry({
    key: PassthroughEncoder for key in DEFAULT_KEYS + ['passthrough']
})


class CategoricalEncoder(Encoder, ABC):
    @classmethod
    def register(cls, name):
        ENCODER_REGISTRY[name] = cls


@register(name='dense')
class CategoricalEmbedEncoder(CategoricalEncoder):

    def __init__(
            self,
            vocab,
            embedding_size=50,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None,
            **kwargs
    ):
        super().__init__()
        logger.debug(' {}'.format(self.name))

        logger.debug('  Embed')
        self.embed = Embed(
            vocab=vocab,
            embedding_size=embedding_size,
            representation='dense',
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=embedding_initializer,
            embedding_regularizer=embedding_regularizer
        )
        self.embedding_size = self.embed.embedding_size

    def call(self, inputs, training=None, mask=None):
        """
            :param inputs: The inputs fed into the encoder.
                   Shape: [batch x 1], type tf.int32

            :param return: embeddings of shape [batch x embed size], type tf.float32
        """
        embedded = self.embed(
            inputs, training=training, mask=mask
        )
        return embedded


@register(name='sparse')
class CategoricalSparseEncoder(CategoricalEncoder):

    def __init__(
            self,
            vocab,
            embedding_size=50,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            dropout=0.0,
            embedding_initializer=None,
            embedding_regularizer=None,
            **kwargs
    ):
        super().__init__()
        logger.debug(' {}'.format(self.name))

        logger.debug('  Embed')
        self.embed = Embed(
            vocab=vocab,
            embedding_size=embedding_size,
            representation='sparse',
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=embedding_initializer,
            embedding_regularizer=embedding_regularizer
        )
        self.embedding_size = self.embed.embedding_size

    def call(self, inputs, training=None, mask=None):
        """
            :param inputs: The inputs fed into the encoder.
                   Shape: [batch x 1], type tf.int32

            :param return: embeddings of shape [batch x embed size], type tf.float32
        """
        embedded = self.embed(
            inputs, training=training, mask=mask
        )
        return embedded



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from jsonschema import validate

from ludwig.combiners.combiners import combiner_registry
from ludwig.features.feature_registries import input_type_registry, output_type_registry

INPUT_FEATURE_TYPES = sorted(list(input_type_registry.keys()))
OUTPUT_FEATURE_TYPES = sorted(list(output_type_registry.keys()))
COMBINER_TYPES = sorted(list(combiner_registry.keys()))


def get_schema():
    schema = {
        'type': 'object',
        'properties': {
            'input_features': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'name': {'type': 'string'},
                        'type': {'type': 'string', 'enum': INPUT_FEATURE_TYPES},
                        'column': {'type': 'string'},
                        'encoder': {'type': 'string'}
                    },
                    'allOf': get_input_encoder_conds() + get_input_preproc_conds(),
                    'required': ['name', 'type'],
                }
            },
            'output_features': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'name': {'type': 'string'},
                        'type': {'type': 'string', 'enum': OUTPUT_FEATURE_TYPES},
                        'column': {'type': 'string'},
                        'decoder': {'type': 'string'}
                    },
                    'allOf': get_output_decoder_conds() + get_output_preproc_conds(),
                    'required': ['name', 'type'],
                }
            },
            'combiner': {
                'type': 'object',
                'properties': {
                    'type': {'type': 'string', 'enum': COMBINER_TYPES},
                },
            },
            'training': {},
            'preprocessing': {},
            'hyperopt': {},
        },
        'required': ['input_features', 'output_features']
    }
    return schema


def get_input_encoder_conds():
    conds = []
    for feature_type in INPUT_FEATURE_TYPES:
        feature_cls = input_type_registry[feature_type]
        encoder_names = list(feature_cls.encoder_registry.keys())
        encoder_cond = create_cond(
            {'type': feature_type},
            {'encoder': {'enum': encoder_names}},
        )
        conds.append(encoder_cond)
    return conds


def get_input_preproc_conds():
    conds = []
    for feature_type in INPUT_FEATURE_TYPES:
        feature_cls = input_type_registry[feature_type]
        preproc_spec = {
            'type': 'object',
            'properties': feature_cls.preprocessing_schema,
            'additionalProperties': False,
        }
        preproc_cond = create_cond(
            {'type': feature_type},
            {'preprocessing': preproc_spec},
        )
        conds.append(preproc_cond)
    return conds


def get_output_decoder_conds():
    conds = []
    for feature_type in OUTPUT_FEATURE_TYPES:
        feature_cls = output_type_registry[feature_type]
        decoder_names = list(feature_cls.decoder_registry.keys())
        decoder_cond = create_cond(
            {'type': feature_type},
            {'decoder': {'enum': decoder_names}},
        )
        conds.append(decoder_cond)
    return conds


def get_output_preproc_conds():
    conds = []
    for feature_type in OUTPUT_FEATURE_TYPES:
        feature_cls = output_type_registry[feature_type]
        preproc_spec = {
            'type': 'object',
            'properties': feature_cls.preprocessing_schema,
            'additionalProperties': False,
        }
        preproc_cond = create_cond(
            {'type': feature_type},
            {'preprocessing': preproc_spec},
        )
        conds.append(preproc_cond)
    return conds


def create_cond(if_pred, then_pred):
    return {
        'if': {
            'properties': {k: {'const': v} for k, v in if_pred.items()}
        },
        'then': {
            'properties': {k: v for k, v in then_pred.items()}
        }
    }


def validate_config(config):
    validate(instance=config, schema=get_schema())



import os.path
import shutil
from io import StringIO

import numpy as np
import pandas as pd
import tensorflow as tf

from ludwig.experiment import full_experiment
from ludwig.predict import full_predict

tf.config.experimental_run_functions_eagerly(True)


# function to generate sequence data like this
#            c1                  y
#            b2                b b
#         c1 b1                c b
#      b1 c1 a2            b c a a
#         e3 a2          e e e a a
def generate_deterministic_sequence(num_records=100):
    in_vocab = [x + str(d) for x in list("abcde") for d in range(1, 4)]

    def generate_output(x):
        letter = x[0]
        repeat = int(x[1])
        return ' '.join(repeat * letter)

    input_list = []
    output_list = []
    for _ in range(num_records):
        n = np.random.randint(1, 5, 1)
        input_seq = np.random.choice(in_vocab, n, replace=True)
        output_seq = [generate_output(x) for x in input_seq]
        input_list.append(' '.join(input_seq))
        output_list.append(' '.join(output_seq))

    train = {
        'in_seq': input_list,
        'out_seq': output_list
    }

    return pd.DataFrame(train)


def dump_key(parms, keys):
    return [key + ": " + str(parms[key]) for key in keys]


shutil.rmtree('./data_to_pred', ignore_errors=True)
os.mkdir('./data_to_pred')
shutil.rmtree('./results', ignore_errors=True)
shutil.rmtree('./results_0', ignore_errors=True)
shutil.rmtree('./results_1', ignore_errors=True)

# From TF1 unit test_experiment.py::test_experiment.py::test_experiment_attention
input_features = [
    {'name': 'in_seq', 'type': 'sequence', 'encoder': 'rnn',
     'cell_type': 'lstm',
     'reduce_output': None}
]
output_features = [
    {'name': 'out_seq', 'type': 'sequence',
     'cell_type': 'lstm',
     'beam_width': 3,
     'decoder': 'generator',
     'attention': 'luong',  # 'luong', #'bahdanau' None
     'reduce_input': None,
     # 'loss': {'type': 'sampled_softmax_cross_entropy'}
     }
]

config = {
    'input_features': input_features,
    'output_features': output_features,
    'training': {'epochs': 15, 'early_stop': 5}
}

args = {
    'config': config,
    'skip_save_processed_input': True,
    'skip_save_progress': True,
    'skip_save_unprocessed_output': True,
    'skip_save_model': False,
    'skip_save_log': False,
    'debug': False
}

np.random.seed(31)  # 13
data_df = generate_deterministic_sequence(128)
test_df = generate_deterministic_sequence(1000)

exp_dir_name = full_experiment(data_df=data_df, **args)

data_df.to_csv('./data_to_pred/train.csv', index=False)

full_predict(
    './results/experiment_run/model',
    data_csv='./data_to_pred/train.csv',
    evaluate_performance=True)

# separate test data set
print(test_df.head(20))
test_df.to_csv('./data_to_pred/my_data.csv', index=False)
full_predict(
    './results/experiment_run/model',
    data_csv='./data_to_pred/my_data.csv',
    evaluate_performance=False)

# check for spurious 0s
# read in predictions
# df = pd.read_csv('./results_1/out_seq_predictions.csv', header=None)
with open('./results_1/out_seq_predictions.csv', 'r') as f:
    buffer = f.read()
buffer = buffer.replace(',', ' ').replace('<PAD>', '-')
df2 = pd.read_csv(StringIO(buffer), header=None)
df2.columns = ['predicted_sequence']

number_predictions = df2['predicted_sequence'].shape[0]
ind_without_end_pad = df2['predicted_sequence'].apply(
    lambda x: False if x[-1] == '-' else True)
number_missing_end_pad = np.sum(ind_without_end_pad)
number_tokens = df2['predicted_sequence'].str.replace(' ', '').str.len()
number_pads = df2['predicted_sequence'].str.count('-')
ind_multiple_pads = number_pads > 1
number_multiple_pads = np.sum(ind_multiple_pads)

try:
    assert not np.any(ind_multiple_pads) and not np.any(ind_without_end_pad)
    print("No spurious 0s or missing 0s")
except AssertionError:
    print("ISSUES FOUND!!!!!!")
    print("number of rows with issues:\n",
          "\tmissing end pads:", number_missing_end_pad, "out of",
          number_predictions, "\n",
          "\tmultiple pads:", number_multiple_pads, "out of",
          number_predictions, "\n"
          )

df2_last = pd.read_csv('./results_1/out_seq_last_predictions.csv', header=None)
df2_last.columns = ['predicted_last']
df3 = pd.concat([test_df, df2, df2_last], axis=1)
pd.options.display.width = None
pd.options.display.max_colwidth = None

print('config:')
print("input features:", dump_key(input_features[0], ['encoder', 'cell_type']))
print("output features", dump_key(output_features[0],
                                  ['decoder', 'cell_type', 'attention',
                                   'beam_width']))
print('training:',
      dump_key(config['training'], ['epochs', 'early_stop']))
print("\nRandom sample of predicted sequences:")
print(df3.sample(n=15, random_state=33))

if number_missing_end_pad > 0:
    print("\n\nsample without end pads:")
    print(df3.loc[ind_without_end_pad].head(10))

if number_multiple_pads > 0:
    print("\n\nsample multiple pads:")
    print(df3.loc[ind_multiple_pads].head(10))



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.datasets.base_dataset import BaseDataset, DEFAULT_CACHE_LOCATION
from ludwig.datasets.mixins.download import UncompressedFileDownloadMixin
from ludwig.datasets.mixins.load import CSVLoadMixin
from ludwig.datasets.mixins.process import MultifileJoinProcessMixin


def load(cache_dir=DEFAULT_CACHE_LOCATION, split=False):
    dataset = GoEmotions(cache_dir=cache_dir)
    return dataset.load(split=split)


class GoEmotions(UncompressedFileDownloadMixin, MultifileJoinProcessMixin,
                 CSVLoadMixin, BaseDataset):
    """The GoEmotions dataset.

    This pulls in an array of mixins for different types of functionality
    which belongs in the workflow for ingesting and transforming training data into a destination
    dataframe that can fit into Ludwig's training API.
    """

    def __init__(self, cache_dir=DEFAULT_CACHE_LOCATION):
        super().__init__(dataset_name="goemotions", cache_dir=cache_dir)

    def read_file(self, filetype, filename, header=None):
        file_df = pd.read_table(os.path.join(self.raw_dataset_path, filename),
                                header=header)
        return file_df

    def process_downloaded_dataset(self):
        super().process_downloaded_dataset()
        # format emotion ids to be a set of emotion ids vs. string
        processed_df = pd.read_csv(os.path.join(self.processed_dataset_path,
                                                self.csv_filename))
        processed_df.columns = ['text', 'emotion_ids', 'comment_id', 'split']
        processed_df['emotion_ids'] = processed_df['emotion_ids'].apply(
            lambda e_id: " ".join(e_id.split(","))
        )
        processed_df.to_csv(
            os.path.join(self.processed_dataset_path, self.csv_filename),
            index=False
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import shutil
import tempfile
from copy import deepcopy

import numpy as np
import pytest
import tensorflow as tf

from ludwig.api import LudwigModel
from ludwig.data.preprocessing import preprocess_for_prediction
from ludwig.globals import TRAIN_SET_METADATA_FILE_NAME
from tests.integration_tests.utils import category_feature, binary_feature, \
    numerical_feature, text_feature, vector_feature, image_feature, \
    audio_feature, timeseries_feature, date_feature, h3_feature, set_feature, \
    bag_feature, LocalTestBackend
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import sequence_feature


@pytest.mark.distributed
@pytest.mark.parametrize('should_load_model', [True, False])
def test_savedmodel(csv_filename, should_load_model):
    #######
    # Setup
    #######
    with tempfile.TemporaryDirectory() as tmpdir:
        dir_path = tmpdir
        data_csv_path = os.path.join(tmpdir, csv_filename)
        image_dest_folder = os.path.join(tmpdir, 'generated_images')
        audio_dest_folder = os.path.join(tmpdir, 'generated_audio')

        # Single sequence input, single category output
        input_features = [
            binary_feature(),
            numerical_feature(),
            category_feature(vocab_size=3),
            sequence_feature(vocab_size=3),
            text_feature(vocab_size=3),
            vector_feature(),
            image_feature(image_dest_folder),
            audio_feature(audio_dest_folder),
            timeseries_feature(),
            date_feature(),
            h3_feature(),
            set_feature(vocab_size=3),
            bag_feature(vocab_size=3),
        ]

        output_features = [
            category_feature(vocab_size=3),
            binary_feature(),
            numerical_feature(),
            sequence_feature(vocab_size=3),
            text_feature(vocab_size=3),
            set_feature(vocab_size=3),
            vector_feature()
        ]

        predictions_column_name = '{}_predictions'.format(
            output_features[0]['name'])

        # Generate test data
        data_csv_path = generate_data(input_features, output_features,
                                      data_csv_path)

        #############
        # Train model
        #############
        backend = LocalTestBackend()
        config = {
            'input_features': input_features,
            'output_features': output_features,
            'training': {'epochs': 2}
        }
        ludwig_model = LudwigModel(config, backend=backend)
        ludwig_model.train(
            dataset=data_csv_path,
            skip_save_training_description=True,
            skip_save_training_statistics=True,
            skip_save_model=True,
            skip_save_progress=True,
            skip_save_log=True,
            skip_save_processed_input=True,
        )

        ###################
        # save Ludwig model
        ###################
        ludwigmodel_path = os.path.join(dir_path, 'ludwigmodel')
        shutil.rmtree(ludwigmodel_path, ignore_errors=True)
        ludwig_model.save(ludwigmodel_path)

        ###################
        # load Ludwig model
        ###################
        if should_load_model:
            ludwig_model = LudwigModel.load(ludwigmodel_path, backend=backend)

        ##############################
        # collect weight tensors names
        ##############################
        original_predictions_df, _ = ludwig_model.predict(
            dataset=data_csv_path)
        original_weights = deepcopy(ludwig_model.model.trainable_variables)

        #################
        # save savedmodel
        #################
        savedmodel_path = os.path.join(dir_path, 'savedmodel')
        shutil.rmtree(savedmodel_path, ignore_errors=True)
        ludwig_model.model.save_savedmodel(savedmodel_path)

        ###################################################
        # load Ludwig model, obtain predictions and weights
        ###################################################
        ludwig_model = LudwigModel.load(ludwigmodel_path, backend=backend)
        loaded_prediction_df, _ = ludwig_model.predict(dataset=data_csv_path)
        loaded_weights = deepcopy(ludwig_model.model.trainable_variables)

        #################################################
        # restore savedmodel, obtain predictions and weights
        #################################################
        training_set_metadata_json_fp = os.path.join(
            ludwigmodel_path,
            TRAIN_SET_METADATA_FILE_NAME
        )

        dataset, training_set_metadata = preprocess_for_prediction(
            ludwig_model.config,
            dataset=data_csv_path,
            training_set_metadata=training_set_metadata_json_fp,
            backend=backend,
        )

        restored_model = tf.saved_model.load(savedmodel_path)

        # Check the outputs for one of the features for correctness
        # Here we choose the first output feature (categorical)
        of_name = list(ludwig_model.model.output_features.keys())[0]

        data_to_predict = {
            name: tf.convert_to_tensor(
                dataset.dataset[feature.proc_column],
                dtype=feature.get_input_dtype()
            )
            for name, feature in ludwig_model.model.input_features.items()
        }

        logits = restored_model(data_to_predict, False, None)

        restored_predictions = tf.argmax(
            logits[of_name]['logits'],
            -1,
            name='predictions_{}'.format(of_name)
        )
        restored_predictions = tf.map_fn(
            lambda idx: training_set_metadata[of_name]['idx2str'][idx],
            restored_predictions,
            dtype=tf.string
        )

        restored_weights = deepcopy(restored_model.trainable_variables)

        #########
        # Cleanup
        #########
        shutil.rmtree(ludwigmodel_path, ignore_errors=True)
        shutil.rmtree(savedmodel_path, ignore_errors=True)

        ###############################################
        # Check if weights and predictions are the same
        ###############################################

        # check for same number of weights as original model
        assert len(original_weights) == len(loaded_weights)
        assert len(original_weights) == len(restored_weights)

        # check to ensure weight valuess match the original model
        loaded_weights_match = np.all(
            [np.all(np.isclose(original_weights[i].numpy(),
                               loaded_weights[i].numpy())) for i in
             range(len(original_weights))]
        )

        original_weights = sorted(original_weights, key=lambda w: w.name)
        restored_weights = sorted(restored_weights, key=lambda w: w.name)

        restored_weights_match = np.all(
            [np.all(np.isclose(original_weights[i].numpy(),
                               restored_weights[i].numpy())) for i in
             range(len(original_weights))]
        )

        assert loaded_weights_match and restored_weights_match

        #  Are predictions identical to original ones?
        loaded_predictions_match = np.all(
            original_predictions_df[predictions_column_name] ==
            loaded_prediction_df[predictions_column_name]
        )

        restored_predictions_match = np.all(
            original_predictions_df[predictions_column_name] ==
            restored_predictions.numpy().astype('str')
        )

        assert loaded_predictions_match and restored_predictions_match



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.data.dataset.pandas import PandasDataset


class PartitionedDataset(object):
    def __init__(self, df, features, data_hdf5_fp):
        self.df = df
        self.features = features
        self.data_hdf5_fp = data_hdf5_fp

    def get(self, col):
        return self.df[col]

    def map_dataset_partitions(self, fn, meta):
        def wrapped(partition):
            dataset = PandasDataset(partition, self.features, self.data_hdf5_fp)
            return fn(dataset)

        return self.df.map_partitions(wrapped, meta=meta)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile

import pytest
import tensorflow as tf

from ludwig.backend import LOCAL_BACKEND
from ludwig.backend.dask import DaskBackend
from ludwig.utils.data_utils import read_parquet

from tests.integration_tests.utils import create_data_set_to_use
from tests.integration_tests.utils import audio_feature
from tests.integration_tests.utils import bag_feature
from tests.integration_tests.utils import binary_feature
from tests.integration_tests.utils import category_feature
from tests.integration_tests.utils import date_feature
from tests.integration_tests.utils import generate_data
from tests.integration_tests.utils import h3_feature
from tests.integration_tests.utils import image_feature
from tests.integration_tests.utils import numerical_feature
from tests.integration_tests.utils import sequence_feature
from tests.integration_tests.utils import set_feature
from tests.integration_tests.utils import spawn
from tests.integration_tests.utils import text_feature
from tests.integration_tests.utils import timeseries_feature
from tests.integration_tests.utils import train_with_backend
from tests.integration_tests.utils import vector_feature


def split(data_parquet):
    data_df = read_parquet(data_parquet, LOCAL_BACKEND.df_engine.df_lib)
    train_df = data_df.sample(frac=0.8)
    test_df = data_df.drop(train_df.index).sample(frac=0.5)
    validation_df = data_df.drop(train_df.index).drop(test_df.index)

    basename, ext = os.path.splitext(data_parquet)
    train_fname = basename + '.train' + ext
    val_fname = basename + '.validation' + ext
    test_fname = basename + '.test' + ext

    train_df.to_parquet(train_fname)
    validation_df.to_parquet(val_fname)
    test_df.to_parquet(test_fname)
    return train_fname, val_fname, test_fname


def run_api_experiment(config, data_parquet, cache_format):
    # Train on Parquet
    dask_backend = DaskBackend(cache_format=cache_format)
    train_with_backend(dask_backend, config, dataset=data_parquet, evaluate=False)

    # Train on DataFrame directly
    data_df = read_parquet(data_parquet, df_lib=dask_backend.df_engine.df_lib)
    train_with_backend(dask_backend, config, dataset=data_df, evaluate=False)


def run_split_api_experiment(config, data_parquet, cache_format):
    backend = DaskBackend(cache_format=cache_format)

    train_fname, val_fname, test_fname = split(data_parquet)

    # Train
    train_with_backend(backend, config,
                       training_set=train_fname,
                       evaluate=False)

    # Train + Validation
    train_with_backend(backend, config,
                       training_set=train_fname,
                       validation_set=val_fname,
                       evaluate=False)

    # Train + Validation + Test
    train_with_backend(backend, config,
                       training_set=train_fname,
                       validation_set=val_fname,
                       test_set=test_fname,
                       evaluate=False)


@spawn
def run_test_parquet(
    input_features,
    output_features,
    num_examples=100,
    run_fn=run_api_experiment,
    expect_error=False,
    cache_format='parquet',
):
    tf.config.experimental_run_functions_eagerly(True)

    config = {
        'input_features': input_features,
        'output_features': output_features,
        'combiner': {'type': 'concat', 'fc_size': 14},
        'training': {'epochs': 2}
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        csv_filename = os.path.join(tmpdir, 'dataset.csv')
        dataset_csv = generate_data(input_features, output_features, csv_filename, num_examples=num_examples)
        dataset_parquet = create_data_set_to_use('parquet', dataset_csv)

        if expect_error:
            with pytest.raises(ValueError):
                run_fn(config, data_parquet=dataset_parquet, cache_format=cache_format)
        else:
            run_fn(config, data_parquet=dataset_parquet, cache_format=cache_format)


@pytest.mark.parametrize('cache_format', ['parquet', 'tfrecord'])
@pytest.mark.distributed
def test_dask_tabular(cache_format):
    input_features = [
        sequence_feature(reduce_output='sum'),
        numerical_feature(normalization='zscore'),
        set_feature(),
        text_feature(),
        binary_feature(),
        bag_feature(),
        vector_feature(),
        h3_feature(),
        date_feature(),
    ]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]
    run_test_parquet(input_features, output_features, cache_format=cache_format)


@pytest.mark.distributed
def test_dask_split():
    input_features = [
        numerical_feature(normalization='zscore'),
        set_feature(),
        binary_feature(),
    ]
    output_features = [category_feature(vocab_size=2, reduce_input='sum')]
    run_test_parquet(input_features, output_features, run_fn=run_split_api_experiment)


@pytest.mark.distributed
def test_dask_timeseries():
    input_features = [timeseries_feature()]
    output_features = [numerical_feature()]
    run_test_parquet(input_features, output_features)


@pytest.mark.distributed
def test_dask_audio():
    with tempfile.TemporaryDirectory() as tmpdir:
        audio_dest_folder = os.path.join(tmpdir, 'generated_audio')
        input_features = [audio_feature(folder=audio_dest_folder)]
        output_features = [binary_feature()]
        run_test_parquet(input_features, output_features, num_examples=25)


@pytest.mark.distributed
def test_dask_lazy_load_audio_error():
    with tempfile.TemporaryDirectory() as tmpdir:
        audio_dest_folder = os.path.join(tmpdir, 'generated_audio')
        input_features = [
            audio_feature(
                folder=audio_dest_folder,
                preprocessing={
                    'in_memory': False,
                }
            )
        ]
        output_features = [binary_feature()]
        run_test_parquet(input_features, output_features, expect_error=True)


@pytest.mark.distributed
def test_dask_image():
    with tempfile.TemporaryDirectory() as tmpdir:
        image_dest_folder = os.path.join(tmpdir, 'generated_images')
        input_features = [
            image_feature(
                folder=image_dest_folder,
                encoder='resnet',
                preprocessing={
                    'in_memory': True,
                    'height': 12,
                    'width': 12,
                    'num_channels': 3,
                    'num_processes': 5
                },
                fc_size=16,
                num_filters=8
            ),
        ]
        output_features = [binary_feature()]
        run_test_parquet(input_features, output_features, num_examples=50)


@pytest.mark.distributed
def test_dask_lazy_load_image_error():
    with tempfile.TemporaryDirectory() as tmpdir:
        image_dest_folder = os.path.join(tmpdir, 'generated_images')
        input_features = [
            image_feature(
                folder=image_dest_folder,
                encoder='resnet',
                preprocessing={
                    'in_memory': False,
                    'height': 12,
                    'width': 12,
                    'num_channels': 3,
                    'num_processes': 5
                },
                fc_size=16,
                num_filters=8
            ),
        ]
        output_features = [binary_feature()]
        run_test_parquet(input_features, output_features, expect_error=True)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib

import numpy as np

from ludwig.constants import PREPROCESSING, TRAINING
from ludwig.data.batcher.random_access import RandomAccessBatcher
from ludwig.data.dataset.base import Dataset
from ludwig.data.sampler import DistributedSampler
from ludwig.utils import data_utils
from ludwig.utils.data_utils import to_numpy_dataset, DATA_TRAIN_HDF5_FP
from ludwig.utils.fs_utils import download_h5
from ludwig.utils.misc_utils import get_proc_features


class PandasDataset(Dataset):
    def __init__(self, dataset, features, data_hdf5_fp):
        self.features = features
        self.data_hdf5_fp = data_hdf5_fp
        self.size = len(dataset)
        self.dataset = to_numpy_dataset(dataset)

    def get(self, proc_column, idx=None):
        if idx is None:
            idx = range(self.size)
        if (self.data_hdf5_fp is None or
                PREPROCESSING not in self.features[proc_column] or
                'in_memory' not in self.features[proc_column][
                    'preprocessing']):
            return self.dataset[proc_column][idx]
        if self.features[proc_column][PREPROCESSING]['in_memory']:
            return self.dataset[proc_column][idx]

        sub_batch = self.dataset[proc_column][idx]

        indices = np.empty((3, len(sub_batch)), dtype=np.int64)
        indices[0, :] = sub_batch
        indices[1, :] = np.arange(len(sub_batch))
        indices = indices[:, np.argsort(indices[0])]

        with download_h5(self.data_hdf5_fp) as h5_file:
            im_data = h5_file[proc_column + '_data'][indices[0, :], :, :]
        indices[2, :] = np.arange(len(sub_batch))
        indices = indices[:, np.argsort(indices[1])]
        return im_data[indices[2, :]]

    def get_dataset(self):
        return self.dataset

    def __len__(self):
        return self.size

    @contextlib.contextmanager
    def initialize_batcher(self, batch_size=128,
                           should_shuffle=True,
                           shuffle_buffer_size=None,
                           seed=0,
                           ignore_last=False,
                           horovod=None):
        sampler = DistributedSampler(len(self),
                                     shuffle=should_shuffle,
                                     seed=seed,
                                     horovod=horovod)
        batcher = RandomAccessBatcher(self,
                                      sampler,
                                      batch_size=batch_size,
                                      ignore_last=ignore_last)
        yield batcher


class PandasDatasetManager(object):
    def __init__(self, backend):
        self.backend = backend

    def create(self, dataset, config, training_set_metadata):
        return PandasDataset(
            dataset,
            get_proc_features(config),
            training_set_metadata.get(DATA_TRAIN_HDF5_FP)
        )

    def create_inference_dataset(self, dataset, tag, config, training_set_metadata):
        return self.create(
            dataset, config, training_set_metadata
        )

    def save(self, cache_path, dataset, config, training_set_metadata, tag):
        data_utils.save_hdf5(cache_path, dataset)
        if tag == TRAINING:
            training_set_metadata[DATA_TRAIN_HDF5_FP] = cache_path
        return dataset

    def can_cache(self, skip_save_processed_input):
        return self.backend.is_coordinator() and \
               not skip_save_processed_input

    @property
    def data_format(self):
        return 'hdf5'
